/* PicasaWeb Downloader v1.1
 * Tomer Gabel, 2007
 * http://www.tomergabel.com
 * 
 * Hacked together in little or no time, please don't take this too seriously.
 * Distributed under the WTFPL license (http://sam.zoy.org/wtfpl/COPYING),
 * which basically means you can do whatever you want with this.
 * 
 * Revision history:
 *  1.0     Initial version
 *  1.1     Added support for AuthKeys
 * 
 */

using System;
using System.IO;
using System.Net;
using System.Text.RegularExpressions;
using System.Xml;

namespace PicasaWebDownloader
{
	/// <summary>
	/// Contains all relevant code for this project.
	/// </summary>
	public class Program
	{
		/// <summary>
		/// Canonizes a PicasaWeb album web/RSS feed URL into its Google API equivalent.
		/// </summary>
		/// <param name="source">Source URL to be canonized</param>
		/// <param name="targetUri">Canonized URI instance</param>
		/// <returns><c>True</c> if the source URL was properly recognized, <c>false</c> otherwise</returns>
		/// <remarks>Currently only handles album and album feed URLs; makes many assumptions (including the
		/// PicasaWeb domain).</remarks>
		static bool canonizePicasaUrl( string source, out Uri targetUri )
		{
			Regex pattern = new Regex(
				@"^(http://picasaweb\.google\.com/" +
				@"((?<username>\w(\w|\.)+)/(?<album>\w(\w|\.)+))" +
				@"|(data/feed/base/user/(?<username>\w(\w|\.)+)/album(?<qual>id)/(?<album>\d+)))" +
                @"(\?.*(?<authkey>authkey\=\w+))?",
                RegexOptions.IgnoreCase | RegexOptions.ExplicitCapture | RegexOptions.IgnorePatternWhitespace );

		    string replacement = 
                "http://picasaweb.google.com/data/feed/api/user/${username}/album${qual}/${album}?kind=photo&${authkey}";

		    Match m = pattern.Match( source );
		    targetUri = m.Success ? new Uri( m.Result( replacement ).TrimEnd( '&' ) ) : null;
		    return m.Success;
		}
		
		/// <summary>
		/// Opens and downloads an image feed.
		/// </summary>
		/// <param name="feedUri">Feed URI</param>
		/// <param name="feed">Creates an XML document containing the downloaded feed</param>
		/// <returns><c>True</c> if the feed was successfully downloaded, <c>false</c> otherwise</returns>
		static bool openFeed( Uri feedUri, out XmlDocument feed )
		{
			feed = null;
			
			Console.Write( "Opening feed..." );
			try
			{
				WebRequest req = WebRequest.Create( feedUri );
				using ( WebResponse resp = req.GetResponse() )
				using ( MemoryStream ms = new MemoryStream() )
				{
					// Download feed
					download( resp, ms, delegate( long offset, long len ) {
					         	Console.Write( "\rDownloading feed ({0}KB)", offset / 1024 ); 
					         } );
					Console.WriteLine( "\rProcessing feed          " );
					
					// Load feed into an XmlDocument instance
					ms.Seek( 0, SeekOrigin.Begin );
					try
					{
						feed = new XmlDocument();
						feed.Load( ms );
						return true;
					}
					catch ( XmlException e )
					{
						Console.WriteLine( "Invalid feed: {0}", e.Message );
						return false;
					}
				}
			}
			catch ( Exception e )
			{
				Console.WriteLine( "\rCan't open feed: {0}", e.Message );
				return false;
			}
		}
		
		/// <summary>
		/// Processes the PicasaWeb API ATOM feed and creates a list of image entries.
		/// </summary>
		/// <param name="feed">XML document containing the PicasaWeb API ATOM feed</param>
		/// <returns>A list of image entries for the specified feed</returns>
		static string[] processEntries( XmlDocument feed )
		{
			// Create a namespace manager (otherwise we can't do squat)
			XmlNamespaceManager xnsm = new XmlNamespaceManager( feed.NameTable );
            xnsm.AddNamespace( "atom", "http://www.w3.org/2005/Atom" );

            // Convoluted XPath expression for extracting the source URLs for all images in the feed
            XmlNodeList entries =
                feed.SelectNodes( "atom:feed/atom:entry[atom:category/@scheme=\"http://schemas.google.com/g/2005#kind\" " +
                                                       "and atom:category/@term=\"http://schemas.google.com/photos/2007#photo\"]" +
                                  "/atom:content/@src", xnsm );
            
            // Create and return the result set
            string[] resultSet = new string[ entries.Count ];
            int i = 0;
            foreach ( XmlNode node in entries )
            	resultSet[ i++ ] = node.Value;
            return resultSet;
		}
		
		/// <summary>
		/// Delegate describing an update request in the download manager
		/// </summary>
		/// <param name="offset">Current position in the downloaded stream</param>
		/// <param name="contentLength">Content length for the downloaded stream</param>
		/// <remarks>Content length is directly derived from the <see cref="T:System.Net.WebResponse"/> and
		/// inherits its quirks.</remarks>
		delegate void DownloadUpdateDelegate( long offset, long contentLength );

		/// <summary>
		/// Streams the entire specified web response, calling a user-defined update function in 16KB increments.
		/// </summary>
		/// <param name="resp">Web response from which to download the file</param>
		/// <param name="target">Target stream</param>
		/// <param name="update">User-defined update delegate for progress report in the UI</param>		
		static void download( WebResponse resp, Stream target, DownloadUpdateDelegate update )
		{
			download( resp, target, 16384, update );
		}
		
		/// <summary>
		/// Streams the entire specified web response, calling a user-defined update function when required.
		/// </summary>
		/// <param name="resp">Web response from which to download the file</param>
		/// <param name="target">Target stream</param>
		/// <param name="updateInterval">Block size (in bytes) between calls to <paramref name="update"/></param>
		/// <param name="update">User-defined update delegate for progress report in the UI</param>		
		static void download( WebResponse resp, Stream target, int updateInterval, DownloadUpdateDelegate update )
		{
			long offset = 0;
			byte[] buffer = new byte[ updateInterval ];
			using ( Stream source = resp.GetResponseStream() )
			{
				int count;
				do
				{
					count = source.Read( buffer, 0, buffer.Length );
					target.Write( buffer, 0, count );
					update( offset += count, resp.ContentLength );
				} while ( count > 0 );
			}
		}
		
		/// <summary>
		/// Downloads a single feed entry.
		/// </summary>
		/// <param name="prefix">Prefix to be shown before the download progress update</param>
		/// <param name="targetFolder">Target folder in which the downloaded file is to be saved</param>
		/// <param name="url">Download URL</param>
		static void downloadFile( string prefix, string targetFolder, string url )
		{
			// Parse URI
			Uri uri;
        	if ( !Uri.TryCreate( url, UriKind.Absolute, out uri ) )
        		Console.WriteLine( "Invalid image URL: " + url );
        	else
        	{
        		string filename = Uri.UnescapeDataString( uri.Segments[ uri.Segments.GetUpperBound( 0 ) ] );
            	string msg      = "\r" + prefix + "{0,5}/{1,5}kb\t{2}";
            	string final    = "\r" + prefix + "       Done  ";
            	Console.Write( msg, '?', '?', filename );
            	
            	try
            	{
            		// Create web request and target file stream
            		HttpWebRequest req = (HttpWebRequest) WebRequest.Create( uri );
            		using ( HttpWebResponse resp = (HttpWebResponse) req.GetResponse() )
            		using ( Stream target = File.Create( Path.Combine( targetFolder, filename ) ) )
            		{
						download( resp, target, delegate( long offset, long len ) {
            			         	Console.Write( msg, offset / 1024, len > 0 ? len / 1024 : '?', filename );
            			         } );
						Console.WriteLine( final );
            		}
            	}
            	catch ( Exception e )
            	{
            		Console.WriteLine( "\tError: " + e.Message );
            	}
        	}
        	
		}

		/// <summary>
		/// Driver script for this project.
		/// </summary>
		/// <param name="args">Command line arguments</param>
        public static void Main( string[] args )
        {
       		Console.WriteLine( "PicasaWeb Downloader by Tomer Gabel, 2007" );
       		Console.WriteLine();
       		
       		// Wrong arguments? Show usage
       		if ( ( args.Length < 1 ) || ( args.Length > 2 ) )
        	{
        		Console.WriteLine( "Usage:" );
        		Console.WriteLine( "\t" + System.Reflection.Assembly.GetExecutingAssembly().GetName().Name + " <url> [target]");
        		Console.WriteLine();
        		Console.WriteLine( "\turl\tURL for the album or its RSS feed" );
        		Console.WriteLine( "\ttarget\tTarget folder for download (e.g. \"C:\\My Pictures\")" );
        		Console.WriteLine( "\t\tIf no target is specified, the current folder is used." );
        		return;
        	}
        	
       		// Attempt to parse and canonize the specified URL
        	Uri feedUri;
        	if ( !canonizePicasaUrl( args[ 0 ], out feedUri ) )
        	{
        		Console.WriteLine( "Unrecognized PicasaWeb URL " + args[ 0 ] );
        		return;
        	}
        	
        	// Set up and, if necessary, create target folder
        	string targetFolder = Directory.GetCurrentDirectory();
        	if ( args.Length > 1 )
        	{
        		targetFolder = args[ 1 ];
        		try 
        		{
        			Directory.CreateDirectory( targetFolder ); 
        		}
        		catch ( Exception e ) 
        		{
        			Console.WriteLine( "Can't create target folder " + targetFolder + ": " + e.Message );
        			return;
        		}
        	}
        	
        	// Open the source feed
        	XmlDocument feed;
        	if ( !openFeed( feedUri, out feed ) )
        	{
        		Console.WriteLine( "Can't open feed " + feedUri.AbsoluteUri );
        		return;
        	}
        	
        	// Process feed, iterate and process entries
        	string[] entries = processEntries( feed );
        	if ( entries.Length > 0 )
        	{
        		int i = 0;
        		foreach ( string url in entries )
        			downloadFile( string.Format( "#{0,3}/{1}\t", ++i, entries.Length ), targetFolder, url );
        	
        		Console.WriteLine();
        		Console.WriteLine( "All done." );
        	}
        	else
        		Console.WriteLine( "No images found in feed!" );
        }
	}
}
